/*
 * ISP driver
 *
 * Author: Kele Bai <kele.bai@amlogic.com>
 *
 * Copyright (C) 2010 Amlogic Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Standard Linux Headers */
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <mach/am_regs.h>

/* Amlogic Headers */
#include <linux/amlogic/tvin/tvin_v4l2.h>

#include "isp_drv.h"

#define DEVICE_NAME "isp"

static void isp_param_show(isp_param_t *parm, int len)
{
	int i=0,j=0;

	for(i=0;i<len;i++){
		if(ISP_U32==parm[i].type || ISP_U16== parm[i].type || ISP_U8== parm[i].type) {
			for(j=0;j<parm[i].length;j++)
				pr_info("%s %s[%d]=0x%x.\n",__func__,parm[i].name,j,*(parm[i].param+j));
		}/* else if(ISP_FLOAT == parm[i].type) {
			for(j=0;j<parm[i].length;j++)
				pr_info("%s %s[%d]=%d.\n",__func__,parm[i].name,j,*(parm[i].param+j));
		}*/
	}
}

static int isp_set_param(isp_param_t *parm,int len,char **buf)
{
	int i=0,j=0;
	for(i=0;i<len;i++) {
		if(!strcmp(parm[i].name,*buf)) {
			if(ISP_U32==parm[i].type || ISP_U8== parm[i].type || ISP_U16== parm[i].type) {
				for(j=0;j<parm[i].length&&*(buf+j+1)!=NULL;j++) {
					*(parm[i].param+j) = simple_strtol(*(buf+j+1),NULL,16);
					pr_info("%s %s[%d]=0x%x.\n",__func__,parm[i].name,j,*(parm[i].param+j));
				}
			} /*else if(ISP_FLOAT == parm[i].type) {
				for(j=0;j<parm[i].length&&*(buf+j+1)!=NULL;j++) {
					sscanf(*(buf+j+1),"%f",(float*)(parm[i].param+j));
					pr_info("%s %s[%d]=%d.\n",__func__,parm[i].name,j,*(parm[i].param+j));
				}
			}*/
			break;
		}
	}
	if(i == len) {
		pr_err("[%s..]%s:the parameter name %s is error.\n",DEVICE_NAME,__func__,buf[0]);
		return 1;
	}
	return 0;
}
isp_param_t isp_param_ae[AE_PARM_NUM]={
		{"ae_algorithm",      NULL,      1,  ISP_U32},
		{"ae_statistics",     NULL,  3,  ISP_U32},
		{"ae_exp",            NULL,         3,  ISP_U32},
		{"ae_ag",             NULL,          3,  ISP_U32},
		{"ae_skip",           NULL,        3,  ISP_U32},
		{"ratio_winl",        NULL,        1,  ISP_U32},
		{"ratio_winr",        NULL,        1,  ISP_U32},
		{"ratio_wint",        NULL,        1,  ISP_U32},
		{"ratio_winb",        NULL,        1,  ISP_U32},
		{"alert_mode",        NULL,        1,  ISP_U32},
		{"tune_mode",         NULL,         1,  ISP_U32},
		{"ratio_r",           NULL,           1,  ISP_U32},
		{"ratio_g",           NULL,           1,  ISP_U32},
		{"ratio_b",           NULL,           1,  ISP_U32},
		{"stepdnr",           NULL,           1,  ISP_U32},
		{"stepdng",           NULL,           1,  ISP_U32},
		{"stepdnb",           NULL,           1,  ISP_U32},
		{"stepup",            NULL,            1,  ISP_U32},
		{"slow_lpfcoef",      NULL,      1,  ISP_U32},
		{"fast_lpfcoef",      NULL,      1,  ISP_U32},
		{"coef_cur",          NULL,       16, ISP_U32},
		{"coef_env",          NULL,       16, ISP_U32},
		{"env_hign",          NULL,          1,  ISP_U32},
		{"env_hign2mid",      NULL,      1,  ISP_U32},
		{"env_low2mid",       NULL,       1,  ISP_U32},
		{"env_low",           NULL,           1,  ISP_U32},
		{"thr_r_high",        NULL,        1,  ISP_U32},
		{"thr_r_mid",         NULL,         1,  ISP_U32},
		{"thr_r_low",         NULL,         1,  ISP_U32},
		{"thr_g_high",        NULL,        1,  ISP_U32},
		{"thr_g_mid",         NULL,         1,  ISP_U32},
		{"thr_g_low",         NULL,         1,  ISP_U32},
		{"thr_b_high",        NULL,        1,  ISP_U32},
		{"thr_b_mid",         NULL,         1,  ISP_U32},
		{"thr_b_low",         NULL,         1,  ISP_U32},
		{"lpftype_high",      NULL,      1,  ISP_U32},
		{"lpftype_mid",       NULL,       1,  ISP_U32},
		{"lpftype_low",       NULL,       1,  ISP_U32},
		{"targethigh",        NULL,        1,  ISP_U32},
		{"targetmid",         NULL,         1,  ISP_U32},
		{"targetlow",         NULL,         1,  ISP_U32},
		{"radium_inner_h",    NULL,    1,  ISP_U32},
		{"radium_outer_h",    NULL,    1,  ISP_U32},
		{"radium_inner_m",    NULL,    1,  ISP_U32},
		{"radium_outer_m",    NULL,    1,  ISP_U32},
		{"radium_inner_l",    NULL,    1,  ISP_U32},
		{"radium_outer_l",    NULL,    1,  ISP_U32},
		{"flash_thr",	      NULL,         1,  ISP_U32},
		{"ratio_histr",	      NULL,       1,  ISP_U32},
		{"ratio_histg",	      NULL,       1,  ISP_U32},
		{"ratio_histb",	      NULL,       1,  ISP_U32},
		{"target_r",	      NULL,          1,  ISP_U32},
		{"target_g",	      NULL,          1,  ISP_U32},
		{"target_b",	      NULL,          1,  ISP_U32},
		{"maxrate_inner",     NULL,     1,  ISP_U32},
		{"maxrate_outer",     NULL,     1,  ISP_U32},
		{"slow_lpfcoef_enh",  NULL,  1,  ISP_U32},
		{"fast_lpfcoef_enh",  NULL,  1,  ISP_U32},
		{"flash_thr_enh",     NULL,     1,  ISP_U32},
		{"ae_ratio_low",	  NULL, 	 1,  ISP_U32},
		{"ae_ratio_low2mid",  NULL,  1,  ISP_U32},
		{"ae_ratio_mid2high", NULL, 1,  ISP_U32},
		{"ae_ratio_high",	  NULL, 	 1,  ISP_U32},
		{"ae_min_diff",	      NULL, 	     1,  ISP_U32},
		{"ae_max_diff",	      NULL, 	     1,  ISP_U32},
		{"reserve",	          NULL, 	     16, ISP_U32},
		{"aet_fmt_gain",	  NULL, 	 1,  ISP_U32},
};
/* ae_param is all u32 if not need to modify here */
void set_ae_parm(xml_algorithm_ae_t *ae_sw,char **parm)
{
	int len = AE_PARM_NUM;
	unsigned int i,*ae_sw_t;
	ae_sw_t = (unsigned int *)ae_sw;
	for(i = 0;i < AE_PARM_NUM;i++){
		isp_param_ae[i].param = ae_sw_t;
		ae_sw_t += isp_param_ae[i].length;
	}
	if(!strcmp(parm[0],"show")){
		isp_param_show((isp_param_t*)&isp_param_ae,len);
	} else {
		isp_set_param((isp_param_t*)&isp_param_ae,len,parm);
	}
}

void set_awb_parm(xml_algorithm_awb_t *awb_sw,char **parm)
{
	int len = AWB_PARM_NUM;
	isp_param_t awb[AWB_PARM_NUM]={
		{"awb_algorithm",  &awb_sw->awb_algorithm,   1,  ISP_U32},
		{"ratio_winl",	   &awb_sw->ratio_winl,		 1,  ISP_U32},
		{"ratio_winr",	   &awb_sw->ratio_winr,		 1,  ISP_U32},
		{"ratio_wint",	   &awb_sw->ratio_wint,		 1,  ISP_U32},
		{"ratio_winb",	   &awb_sw->ratio_winb,		 1,  ISP_U32},
		{"ratio_rgb",      &awb_sw->ratio_rgb,       1,  ISP_U32},
		{"ratio_yh",       &awb_sw->ratio_yh,        1,  ISP_U32},
		{"ratio_ym",       &awb_sw->ratio_ym,        1,  ISP_U32},
		{"ratio_yl",       &awb_sw->ratio_yl,        1,  ISP_U32},
		{"yyh",            &awb_sw->yyh,             1,  ISP_U32},
		{"yym",            &awb_sw->yym,             1,  ISP_U32},
		{"yyl",            &awb_sw->yyl,             1,  ISP_U32},
		{"coef_r",         &awb_sw->coef_r[0],       4,  ISP_U32},
		{"coef_g",         &awb_sw->coef_g[0],       4,  ISP_U32},
		{"coef_b",         &awb_sw->coef_b[0],       4,  ISP_U32},
		{"inner_rg",       &awb_sw->inner_rg,        1,  ISP_U32},
		{"outer_rg",       &awb_sw->outer_rg,        1,  ISP_U32},
		{"inner_bg",       &awb_sw->inner_bg,        1,  ISP_U32},
		{"outer_bg",       &awb_sw->outer_bg,        1,  ISP_U32},
		{"r_max",          &awb_sw->r_max,           1,  ISP_U32},
		{"r_min",          &awb_sw->r_min,           1,  ISP_U32},
		{"b_max",          &awb_sw->b_max,           1,  ISP_U32},
		{"b_min",          &awb_sw->b_min,           1,  ISP_U32},
		{"thr_gb_h",	   &awb_sw->thr_gb_h,	     1,  ISP_U32},
		{"thr_gb_m",	   &awb_sw->thr_gb_m,	     1,  ISP_U32},
		{"thr_gb_l",	   &awb_sw->thr_gb_l,	     1,  ISP_U32},
		{"thr_gr_h",	   &awb_sw->thr_gr_h,	     1,  ISP_U32},
		{"thr_gr_m",	   &awb_sw->thr_gr_m,	     1,  ISP_U32},
		{"thr_gr_l",	   &awb_sw->thr_gr_l,	     1,  ISP_U32},
		{"thr_br_h",	   &awb_sw->thr_br_h,	     1,  ISP_U32},
		{"thr_br_m",	   &awb_sw->thr_br_m,	     1,  ISP_U32},
		{"thr_br_l",	   &awb_sw->thr_br_l,	     1,  ISP_U32},
		{"thr_du_h",	   &awb_sw->thr_du_h,	     1,  ISP_U32},
		{"thr_du_m",	   &awb_sw->thr_du_m,	     1,  ISP_U32},
		{"thr_du_l",	   &awb_sw->thr_du_l,	     1,  ISP_U32},
		{"thr_dv_h",	   &awb_sw->thr_dv_h,	     1,  ISP_U32},
		{"thr_dv_m",	   &awb_sw->thr_dv_m,	     1,  ISP_U32},
		{"thr_dv_l",	   &awb_sw->thr_dv_l,	     1,  ISP_U32},
		{"thr_yh_h",	   &awb_sw->thr_yh_h,	     1,  ISP_U32},
		{"thr_yh_m",	   &awb_sw->thr_yh_m,	     1,  ISP_U32},
		{"thr_yh_l",	   &awb_sw->thr_yh_l,	     1,  ISP_U32},
		{"thr_yl_h",	   &awb_sw->thr_yl_h,	     1,  ISP_U32},
		{"thr_yl_m",	   &awb_sw->thr_yl_m,	     1,  ISP_U32},
		{"thr_yl_l",	   &awb_sw->thr_yl_l,	     1,  ISP_U32},
	        {"ratio_yuv",      &awb_sw->ratio_yuv,       1,  ISP_U32},
		{"slow_lpfcoef",   &awb_sw->slow_lpfcoef,    1,  ISP_U32},
		{"fast_lpfcoef",   &awb_sw->fast_lpfcoef,    1,  ISP_U32},
		{"outer",          &awb_sw->outer,           1,  ISP_U32},
		{"inner",          &awb_sw->inner,           1,  ISP_U32},
		{"rw_limith",      &awb_sw->rw_limith,       1,  ISP_U32},
		{"rw_limitl",      &awb_sw->rw_limitl,       1,  ISP_U32},
		{"gw_limith",      &awb_sw->gw_limith,       1,  ISP_U32},
		{"gw_limitl",      &awb_sw->gw_limitl,       1,  ISP_U32},
		{"bw_limith",      &awb_sw->bw_limith,       1,  ISP_U32},
		{"bw_limitl",      &awb_sw->bw_limitl,       1,  ISP_U32},
		{"thr_u",          &awb_sw->thr_u[0],        20, ISP_U32},
		{"thr_v",          &awb_sw->thr_v[0],        20, ISP_U32},
		{"reserve",        &awb_sw->reserve[0],      16, ISP_U32},
	};

	if(!strcmp(parm[0],"show")){
		isp_param_show((isp_param_t*)&awb,len);
	} else {
		isp_set_param((isp_param_t*)&awb,len,parm);
	}
}

void set_af_parm(xml_algorithm_af_t *af_sw,char **parm)
{
	int len = AF_PARM_NUM;
	isp_param_t af[AF_PARM_NUM]={
		{"enter_static_ratio",    &af_sw->enter_static_ratio,      1, ISP_U32},
		{"detect_step_cnt",       &af_sw->detect_step_cnt,         1, ISP_U32},
		{"ave_vdc_thr",		  &af_sw->ave_vdc_thr,		   		   1, ISP_U32},
		{"win_ratio", 	  	  &af_sw->win_ratio,		   1, ISP_U32},
		{"step",      		  (unsigned int *)&af_sw->step,           FOCUS_GRIDS, ISP_U32},
		{"valid_step_cnt",        &af_sw->valid_step_cnt,          1, ISP_U32},
		{"jump_offset",           &af_sw->jump_offset,             1, ISP_U32},
		{"field_delay",           &af_sw->field_delay,             1, ISP_U32},
		{"x",                     &af_sw->x,      			       1, ISP_U32},
		{"y",                     &af_sw->y,      			       1, ISP_U32},
		{"radius_ratio",          &af_sw->radius_ratio,     		       1, ISP_U32},
		{"radius_ratio",          &af_sw->hillside_fall,     		       1, ISP_U32},
	};

	if(!strcmp(parm[0],"show")){
		isp_param_show((isp_param_t*)&af,len);
	} else {
		isp_set_param((isp_param_t*)&af,len,parm);
	}
}

void set_cap_parm(struct xml_capture_s *cap_sw,char **parm)
{
	int len = CAP_PARM_NUM;
	isp_param_t cap[CAP_PARM_NUM]={
		{"ae_try_max_cnt",   &cap_sw->ae_try_max_cnt,      1, ISP_U32},
		{"af_mode",          &cap_sw->af_mode,      	   1, ISP_U32},
		{"sigle_count",      &cap_sw->sigle_count,         1, ISP_U32},
		{"skip_step",        &cap_sw->skip_step,      	   1, ISP_U32},
		{"multi_capture_num",&cap_sw->multi_capture_num,   1, ISP_U32},
		{"eyetime",          &cap_sw->eyetime,      	   1, ISP_U32},
		{"pretime",          &cap_sw->pretime,      	   1, ISP_U32},
		{"postime",          &cap_sw->postime,      	   1, ISP_U32},
	};

	if(!strcmp(parm[0],"show")){
		isp_param_show((isp_param_t*)&cap,len);
	} else {
		isp_set_param((isp_param_t*)&cap,len,parm);
	}

}

void set_wave_parm(struct wave_s *wave,char **parm)
{
	int len = WAVE_PARM_NUM;
	isp_param_t wav[WAVE_PARM_NUM]={
		{"torch_rising_time",      	&wave->torch_rising_time,      	1, ISP_U32},
		{"flash_rising_time",           &wave->flash_rising_time,      	1, ISP_U32},
		{"torch_flash_ratio",           &wave->torch_flash_ratio,      	1, ISP_U32},
		{"wave_clock_div",              &wave->wave_clock_div,          1, ISP_U32},
		{"pulse_init_time",             &wave->pulse_init_time,         1, ISP_U32},
		{"pulse_high_time",             &wave->pulse_high_time,      	1, ISP_U32},
		{"pulse_low_time",              &wave->pulse_low_time,          1, ISP_U32},
		{"time_to_latch",               &wave->time_to_latch,      	1, ISP_U32},
		{"latch_time",                  &wave->latch_time,      	1, ISP_U32},
		{"latch_time_timeout",          &wave->latch_time_timeout,      1, ISP_U32},
		{"time_to_off",                 &wave->time_to_off,      	1, ISP_U32},
		{"pulse_qty_max",               &wave->pulse_qty_max,      	1, ISP_U32},
	};

	if(!strcmp(parm[0],"show")){
		isp_param_show((isp_param_t*)&wav,len);
	} else {
		isp_set_param((isp_param_t*)&wav,len,parm);
	}

}
static unsigned int gamma[10][257] =
	{
		// curve_ratio = 1.0
		{
			0x00000000,
			0x00000004,
			0x00000008,
			0x0000000C,
			0x00000010,
			0x00000014,
			0x00000018,
			0x0000001C,
			0x00000020,
			0x00000024,
			0x00000028,
			0x0000002C,
			0x00000030,
			0x00000034,
			0x00000038,
			0x0000003C,
			0x00000040,
			0x00000044,
			0x00000048,
			0x0000004C,
			0x00000050,
			0x00000054,
			0x00000058,
			0x0000005C,
			0x00000060,
			0x00000064,
			0x00000068,
			0x0000006C,
			0x00000070,
			0x00000074,
			0x00000078,
			0x0000007C,
			0x00000080,
			0x00000084,
			0x00000088,
			0x0000008C,
			0x00000090,
			0x00000094,
			0x00000098,
			0x0000009C,
			0x000000A0,
			0x000000A4,
			0x000000A8,
			0x000000AC,
			0x000000B0,
			0x000000B4,
			0x000000B8,
			0x000000BC,
			0x000000C0,
			0x000000C4,
			0x000000C8,
			0x000000CC,
			0x000000D0,
			0x000000D4,
			0x000000D8,
			0x000000DC,
			0x000000E0,
			0x000000E4,
			0x000000E8,
			0x000000EC,
			0x000000F0,
			0x000000F4,
			0x000000F8,
			0x000000FC,
			0x00000100,
			0x00000104,
			0x00000108,
			0x0000010C,
			0x00000110,
			0x00000114,
			0x00000118,
			0x0000011C,
			0x00000120,
			0x00000124,
			0x00000128,
			0x0000012C,
			0x00000130,
			0x00000134,
			0x00000138,
			0x0000013C,
			0x00000140,
			0x00000144,
			0x00000148,
			0x0000014C,
			0x00000150,
			0x00000154,
			0x00000158,
			0x0000015C,
			0x00000160,
			0x00000164,
			0x00000168,
			0x0000016C,
			0x00000170,
			0x00000174,
			0x00000178,
			0x0000017C,
			0x00000180,
			0x00000184,
			0x00000188,
			0x0000018C,
			0x00000190,
			0x00000194,
			0x00000198,
			0x0000019C,
			0x000001A0,
			0x000001A4,
			0x000001A8,
			0x000001AC,
			0x000001B0,
			0x000001B4,
			0x000001B8,
			0x000001BC,
			0x000001C0,
			0x000001C4,
			0x000001C8,
			0x000001CC,
			0x000001D0,
			0x000001D4,
			0x000001D8,
			0x000001DC,
			0x000001E0,
			0x000001E4,
			0x000001E8,
			0x000001EC,
			0x000001F0,
			0x000001F4,
			0x000001F8,
			0x000001FC,
			0x00000200,
			0x00000204,
			0x00000208,
			0x0000020C,
			0x00000210,
			0x00000214,
			0x00000218,
			0x0000021C,
			0x00000220,
			0x00000224,
			0x00000228,
			0x0000022C,
			0x00000230,
			0x00000234,
			0x00000238,
			0x0000023C,
			0x00000240,
			0x00000244,
			0x00000248,
			0x0000024C,
			0x00000250,
			0x00000254,
			0x00000258,
			0x0000025C,
			0x00000260,
			0x00000264,
			0x00000268,
			0x0000026C,
			0x00000270,
			0x00000274,
			0x00000278,
			0x0000027C,
			0x00000280,
			0x00000284,
			0x00000288,
			0x0000028C,
			0x00000290,
			0x00000294,
			0x00000298,
			0x0000029C,
			0x000002A0,
			0x000002A4,
			0x000002A8,
			0x000002AC,
			0x000002B0,
			0x000002B4,
			0x000002B8,
			0x000002BC,
			0x000002C0,
			0x000002C4,
			0x000002C8,
			0x000002CC,
			0x000002D0,
			0x000002D4,
			0x000002D8,
			0x000002DC,
			0x000002E0,
			0x000002E4,
			0x000002E8,
			0x000002EC,
			0x000002F0,
			0x000002F4,
			0x000002F8,
			0x000002FC,
			0x00000300,
			0x00000304,
			0x00000308,
			0x0000030C,
			0x00000310,
			0x00000314,
			0x00000318,
			0x0000031C,
			0x00000320,
			0x00000324,
			0x00000328,
			0x0000032C,
			0x00000330,
			0x00000334,
			0x00000338,
			0x0000033C,
			0x00000340,
			0x00000344,
			0x00000348,
			0x0000034C,
			0x00000350,
			0x00000354,
			0x00000358,
			0x0000035C,
			0x00000360,
			0x00000364,
			0x00000368,
			0x0000036C,
			0x00000370,
			0x00000374,
			0x00000378,
			0x0000037C,
			0x00000380,
			0x00000384,
			0x00000388,
			0x0000038C,
			0x00000390,
			0x00000394,
			0x00000398,
			0x0000039C,
			0x000003A0,
			0x000003A4,
			0x000003A8,
			0x000003AC,
			0x000003B0,
			0x000003B4,
			0x000003B8,
			0x000003BC,
			0x000003C0,
			0x000003C4,
			0x000003C8,
			0x000003CC,
			0x000003D0,
			0x000003D4,
			0x000003D8,
			0x000003DC,
			0x000003E0,
			0x000003E4,
			0x000003E8,
			0x000003EC,
			0x000003F0,
			0x000003F4,
			0x000003F8,
			0x000003FC,
			0x000003FF,
		},
		// curve_ratio = 0.1
		{
			0x00000000,
			0x0000024C,
			0x00000276,
			0x00000290,
			0x000002A4,
			0x000002B3,
			0x000002C0,
			0x000002CA,
			0x000002D4,
			0x000002DD,
			0x000002E4,
			0x000002EC,
			0x000002F2,
			0x000002F8,
			0x000002FE,
			0x00000303,
			0x00000308,
			0x0000030D,
			0x00000311,
			0x00000315,
			0x0000031A,
			0x0000031D,
			0x00000321,
			0x00000325,
			0x00000328,
			0x0000032B,
			0x0000032F,
			0x00000332,
			0x00000335,
			0x00000338,
			0x0000033A,
			0x0000033D,
			0x00000340,
			0x00000342,
			0x00000345,
			0x00000347,
			0x0000034A,
			0x0000034C,
			0x0000034E,
			0x00000350,
			0x00000353,
			0x00000355,
			0x00000357,
			0x00000359,
			0x0000035B,
			0x0000035D,
			0x0000035E,
			0x00000360,
			0x00000362,
			0x00000364,
			0x00000366,
			0x00000367,
			0x00000369,
			0x0000036B,
			0x0000036C,
			0x0000036E,
			0x00000370,
			0x00000371,
			0x00000373,
			0x00000374,
			0x00000376,
			0x00000377,
			0x00000379,
			0x0000037A,
			0x0000037B,
			0x0000037D,
			0x0000037E,
			0x00000380,
			0x00000381,
			0x00000382,
			0x00000383,
			0x00000385,
			0x00000386,
			0x00000387,
			0x00000388,
			0x0000038A,
			0x0000038B,
			0x0000038C,
			0x0000038D,
			0x0000038E,
			0x00000390,
			0x00000391,
			0x00000392,
			0x00000393,
			0x00000394,
			0x00000395,
			0x00000396,
			0x00000397,
			0x00000398,
			0x00000399,
			0x0000039A,
			0x0000039B,
			0x0000039C,
			0x0000039D,
			0x0000039E,
			0x0000039F,
			0x000003A0,
			0x000003A1,
			0x000003A2,
			0x000003A3,
			0x000003A4,
			0x000003A5,
			0x000003A6,
			0x000003A7,
			0x000003A8,
			0x000003A9,
			0x000003AA,
			0x000003AA,
			0x000003AB,
			0x000003AC,
			0x000003AD,
			0x000003AE,
			0x000003AF,
			0x000003B0,
			0x000003B0,
			0x000003B1,
			0x000003B2,
			0x000003B3,
			0x000003B4,
			0x000003B4,
			0x000003B5,
			0x000003B6,
			0x000003B7,
			0x000003B8,
			0x000003B8,
			0x000003B9,
			0x000003BA,
			0x000003BB,
			0x000003BB,
			0x000003BC,
			0x000003BD,
			0x000003BE,
			0x000003BE,
			0x000003BF,
			0x000003C0,
			0x000003C1,
			0x000003C1,
			0x000003C2,
			0x000003C3,
			0x000003C3,
			0x000003C4,
			0x000003C5,
			0x000003C5,
			0x000003C6,
			0x000003C7,
			0x000003C7,
			0x000003C8,
			0x000003C9,
			0x000003C9,
			0x000003CA,
			0x000003CB,
			0x000003CB,
			0x000003CC,
			0x000003CD,
			0x000003CD,
			0x000003CE,
			0x000003CF,
			0x000003CF,
			0x000003D0,
			0x000003D0,
			0x000003D1,
			0x000003D2,
			0x000003D2,
			0x000003D3,
			0x000003D3,
			0x000003D4,
			0x000003D5,
			0x000003D5,
			0x000003D6,
			0x000003D6,
			0x000003D7,
			0x000003D8,
			0x000003D8,
			0x000003D9,
			0x000003D9,
			0x000003DA,
			0x000003DA,
			0x000003DB,
			0x000003DB,
			0x000003DC,
			0x000003DD,
			0x000003DD,
			0x000003DE,
			0x000003DE,
			0x000003DF,
			0x000003DF,
			0x000003E0,
			0x000003E0,
			0x000003E1,
			0x000003E1,
			0x000003E2,
			0x000003E2,
			0x000003E3,
			0x000003E3,
			0x000003E4,
			0x000003E5,
			0x000003E5,
			0x000003E6,
			0x000003E6,
			0x000003E7,
			0x000003E7,
			0x000003E8,
			0x000003E8,
			0x000003E9,
			0x000003E9,
			0x000003EA,
			0x000003EA,
			0x000003EA,
			0x000003EB,
			0x000003EB,
			0x000003EC,
			0x000003EC,
			0x000003ED,
			0x000003ED,
			0x000003EE,
			0x000003EE,
			0x000003EF,
			0x000003EF,
			0x000003F0,
			0x000003F0,
			0x000003F1,
			0x000003F1,
			0x000003F2,
			0x000003F2,
			0x000003F2,
			0x000003F3,
			0x000003F3,
			0x000003F4,
			0x000003F4,
			0x000003F5,
			0x000003F5,
			0x000003F6,
			0x000003F6,
			0x000003F6,
			0x000003F7,
			0x000003F7,
			0x000003F8,
			0x000003F8,
			0x000003F9,
			0x000003F9,
			0x000003F9,
			0x000003FA,
			0x000003FA,
			0x000003FB,
			0x000003FB,
			0x000003FC,
			0x000003FC,
			0x000003FC,
			0x000003FD,
			0x000003FD,
			0x000003FE,
			0x000003FE,
			0x000003FE,
			0x000003FF,
			0x000003FF,
			0x000003FF,
			0x000003FF,
		},
		// curve_ratio = 0.2
		{
			0x00000000,
			0x00000152,
			0x00000184,
			0x000001A5,
			0x000001BE,
			0x000001D2,
			0x000001E3,
			0x000001F3,
			0x00000200,
			0x0000020C,
			0x00000217,
			0x00000222,
			0x0000022B,
			0x00000234,
			0x0000023D,
			0x00000245,
			0x0000024C,
			0x00000253,
			0x0000025A,
			0x00000261,
			0x00000267,
			0x0000026D,
			0x00000273,
			0x00000278,
			0x0000027E,
			0x00000283,
			0x00000288,
			0x0000028D,
			0x00000292,
			0x00000296,
			0x0000029B,
			0x0000029F,
			0x000002A4,
			0x000002A8,
			0x000002AC,
			0x000002B0,
			0x000002B4,
			0x000002B7,
			0x000002BB,
			0x000002BF,
			0x000002C2,
			0x000002C6,
			0x000002C9,
			0x000002CD,
			0x000002D0,
			0x000002D3,
			0x000002D6,
			0x000002DA,
			0x000002DD,
			0x000002E0,
			0x000002E3,
			0x000002E6,
			0x000002E8,
			0x000002EB,
			0x000002EE,
			0x000002F1,
			0x000002F4,
			0x000002F6,
			0x000002F9,
			0x000002FC,
			0x000002FE,
			0x00000301,
			0x00000303,
			0x00000306,
			0x00000308,
			0x0000030A,
			0x0000030D,
			0x0000030F,
			0x00000312,
			0x00000314,
			0x00000316,
			0x00000318,
			0x0000031B,
			0x0000031D,
			0x0000031F,
			0x00000321,
			0x00000323,
			0x00000325,
			0x00000327,
			0x00000329,
			0x0000032B,
			0x0000032D,
			0x0000032F,
			0x00000331,
			0x00000333,
			0x00000335,
			0x00000337,
			0x00000339,
			0x0000033B,
			0x0000033D,
			0x0000033F,
			0x00000341,
			0x00000342,
			0x00000344,
			0x00000346,
			0x00000348,
			0x0000034A,
			0x0000034B,
			0x0000034D,
			0x0000034F,
			0x00000351,
			0x00000352,
			0x00000354,
			0x00000356,
			0x00000357,
			0x00000359,
			0x0000035A,
			0x0000035C,
			0x0000035E,
			0x0000035F,
			0x00000361,
			0x00000362,
			0x00000364,
			0x00000365,
			0x00000367,
			0x00000369,
			0x0000036A,
			0x0000036C,
			0x0000036D,
			0x0000036F,
			0x00000370,
			0x00000371,
			0x00000373,
			0x00000374,
			0x00000376,
			0x00000377,
			0x00000379,
			0x0000037A,
			0x0000037B,
			0x0000037D,
			0x0000037E,
			0x00000380,
			0x00000381,
			0x00000382,
			0x00000384,
			0x00000385,
			0x00000386,
			0x00000388,
			0x00000389,
			0x0000038A,
			0x0000038C,
			0x0000038D,
			0x0000038E,
			0x0000038F,
			0x00000391,
			0x00000392,
			0x00000393,
			0x00000394,
			0x00000396,
			0x00000397,
			0x00000398,
			0x00000399,
			0x0000039B,
			0x0000039C,
			0x0000039D,
			0x0000039E,
			0x0000039F,
			0x000003A1,
			0x000003A2,
			0x000003A3,
			0x000003A4,
			0x000003A5,
			0x000003A6,
			0x000003A8,
			0x000003A9,
			0x000003AA,
			0x000003AB,
			0x000003AC,
			0x000003AD,
			0x000003AE,
			0x000003AF,
			0x000003B1,
			0x000003B2,
			0x000003B3,
			0x000003B4,
			0x000003B5,
			0x000003B6,
			0x000003B7,
			0x000003B8,
			0x000003B9,
			0x000003BA,
			0x000003BB,
			0x000003BC,
			0x000003BE,
			0x000003BF,
			0x000003C0,
			0x000003C1,
			0x000003C2,
			0x000003C3,
			0x000003C4,
			0x000003C5,
			0x000003C6,
			0x000003C7,
			0x000003C8,
			0x000003C9,
			0x000003CA,
			0x000003CB,
			0x000003CC,
			0x000003CD,
			0x000003CE,
			0x000003CF,
			0x000003D0,
			0x000003D1,
			0x000003D2,
			0x000003D3,
			0x000003D3,
			0x000003D4,
			0x000003D5,
			0x000003D6,
			0x000003D7,
			0x000003D8,
			0x000003D9,
			0x000003DA,
			0x000003DB,
			0x000003DC,
			0x000003DD,
			0x000003DE,
			0x000003DF,
			0x000003E0,
			0x000003E1,
			0x000003E1,
			0x000003E2,
			0x000003E3,
			0x000003E4,
			0x000003E5,
			0x000003E6,
			0x000003E7,
			0x000003E8,
			0x000003E9,
			0x000003E9,
			0x000003EA,
			0x000003EB,
			0x000003EC,
			0x000003ED,
			0x000003EE,
			0x000003EF,
			0x000003EF,
			0x000003F0,
			0x000003F1,
			0x000003F2,
			0x000003F3,
			0x000003F4,
			0x000003F5,
			0x000003F5,
			0x000003F6,
			0x000003F7,
			0x000003F8,
			0x000003F9,
			0x000003FA,
			0x000003FA,
			0x000003FB,
			0x000003FC,
			0x000003FD,
			0x000003FE,
			0x000003FE,
			0x000003FF,
			0x000003FF,
		},
		// curve_ratio = 0.3
		{
			0x00000000,
			0x000000C2,
			0x000000EF,
			0x0000010E,
			0x00000126,
			0x0000013A,
			0x0000014C,
			0x0000015C,
			0x0000016A,
			0x00000177,
			0x00000183,
			0x0000018E,
			0x00000199,
			0x000001A3,
			0x000001AC,
			0x000001B5,
			0x000001BE,
			0x000001C6,
			0x000001CE,
			0x000001D5,
			0x000001DD,
			0x000001E4,
			0x000001EA,
			0x000001F1,
			0x000001F7,
			0x000001FE,
			0x00000204,
			0x00000209,
			0x0000020F,
			0x00000215,
			0x0000021A,
			0x00000220,
			0x00000225,
			0x0000022A,
			0x0000022F,
			0x00000234,
			0x00000238,
			0x0000023D,
			0x00000242,
			0x00000246,
			0x0000024B,
			0x0000024F,
			0x00000253,
			0x00000258,
			0x0000025C,
			0x00000260,
			0x00000264,
			0x00000268,
			0x0000026C,
			0x00000270,
			0x00000273,
			0x00000277,
			0x0000027B,
			0x0000027E,
			0x00000282,
			0x00000286,
			0x00000289,
			0x0000028D,
			0x00000290,
			0x00000293,
			0x00000297,
			0x0000029A,
			0x0000029D,
			0x000002A0,
			0x000002A4,
			0x000002A7,
			0x000002AA,
			0x000002AD,
			0x000002B0,
			0x000002B3,
			0x000002B6,
			0x000002B9,
			0x000002BC,
			0x000002BF,
			0x000002C2,
			0x000002C5,
			0x000002C7,
			0x000002CA,
			0x000002CD,
			0x000002D0,
			0x000002D2,
			0x000002D5,
			0x000002D8,
			0x000002DA,
			0x000002DD,
			0x000002E0,
			0x000002E2,
			0x000002E5,
			0x000002E7,
			0x000002EA,
			0x000002EC,
			0x000002EF,
			0x000002F1,
			0x000002F4,
			0x000002F6,
			0x000002F9,
			0x000002FB,
			0x000002FD,
			0x00000300,
			0x00000302,
			0x00000304,
			0x00000307,
			0x00000309,
			0x0000030B,
			0x0000030E,
			0x00000310,
			0x00000312,
			0x00000314,
			0x00000316,
			0x00000319,
			0x0000031B,
			0x0000031D,
			0x0000031F,
			0x00000321,
			0x00000323,
			0x00000325,
			0x00000328,
			0x0000032A,
			0x0000032C,
			0x0000032E,
			0x00000330,
			0x00000332,
			0x00000334,
			0x00000336,
			0x00000338,
			0x0000033A,
			0x0000033C,
			0x0000033E,
			0x00000340,
			0x00000342,
			0x00000344,
			0x00000346,
			0x00000347,
			0x00000349,
			0x0000034B,
			0x0000034D,
			0x0000034F,
			0x00000351,
			0x00000353,
			0x00000355,
			0x00000356,
			0x00000358,
			0x0000035A,
			0x0000035C,
			0x0000035E,
			0x0000035F,
			0x00000361,
			0x00000363,
			0x00000365,
			0x00000367,
			0x00000368,
			0x0000036A,
			0x0000036C,
			0x0000036D,
			0x0000036F,
			0x00000371,
			0x00000373,
			0x00000374,
			0x00000376,
			0x00000378,
			0x00000379,
			0x0000037B,
			0x0000037D,
			0x0000037E,
			0x00000380,
			0x00000382,
			0x00000383,
			0x00000385,
			0x00000386,
			0x00000388,
			0x0000038A,
			0x0000038B,
			0x0000038D,
			0x0000038E,
			0x00000390,
			0x00000392,
			0x00000393,
			0x00000395,
			0x00000396,
			0x00000398,
			0x00000399,
			0x0000039B,
			0x0000039C,
			0x0000039E,
			0x0000039F,
			0x000003A1,
			0x000003A2,
			0x000003A4,
			0x000003A5,
			0x000003A7,
			0x000003A8,
			0x000003AA,
			0x000003AB,
			0x000003AD,
			0x000003AE,
			0x000003B0,
			0x000003B1,
			0x000003B3,
			0x000003B4,
			0x000003B5,
			0x000003B7,
			0x000003B8,
			0x000003BA,
			0x000003BB,
			0x000003BD,
			0x000003BE,
			0x000003BF,
			0x000003C1,
			0x000003C2,
			0x000003C4,
			0x000003C5,
			0x000003C6,
			0x000003C8,
			0x000003C9,
			0x000003CA,
			0x000003CC,
			0x000003CD,
			0x000003CE,
			0x000003D0,
			0x000003D1,
			0x000003D2,
			0x000003D4,
			0x000003D5,
			0x000003D6,
			0x000003D8,
			0x000003D9,
			0x000003DA,
			0x000003DC,
			0x000003DD,
			0x000003DE,
			0x000003E0,
			0x000003E1,
			0x000003E2,
			0x000003E3,
			0x000003E5,
			0x000003E6,
			0x000003E7,
			0x000003E9,
			0x000003EA,
			0x000003EB,
			0x000003EC,
			0x000003EE,
			0x000003EF,
			0x000003F0,
			0x000003F1,
			0x000003F3,
			0x000003F4,
			0x000003F5,
			0x000003F6,
			0x000003F8,
			0x000003F9,
			0x000003FA,
			0x000003FB,
			0x000003FC,
			0x000003FE,
			0x000003FF,
			0x000003FF,
		},
		// curve_ratio = 0.4
		{
			0x00000000,
			0x0000006F,
			0x00000093,
			0x000000AD,
			0x000000C2,
			0x000000D4,
			0x000000E4,
			0x000000F3,
			0x00000100,
			0x0000010C,
			0x00000118,
			0x00000123,
			0x0000012D,
			0x00000137,
			0x00000140,
			0x00000149,
			0x00000152,
			0x0000015A,
			0x00000162,
			0x0000016A,
			0x00000171,
			0x00000179,
			0x00000180,
			0x00000187,
			0x0000018D,
			0x00000194,
			0x0000019A,
			0x000001A0,
			0x000001A7,
			0x000001AD,
			0x000001B2,
			0x000001B8,
			0x000001BE,
			0x000001C3,
			0x000001C9,
			0x000001CE,
			0x000001D3,
			0x000001D8,
			0x000001DD,
			0x000001E2,
			0x000001E7,
			0x000001EC,
			0x000001F1,
			0x000001F6,
			0x000001FA,
			0x000001FF,
			0x00000203,
			0x00000208,
			0x0000020C,
			0x00000211,
			0x00000215,
			0x00000219,
			0x0000021D,
			0x00000221,
			0x00000225,
			0x0000022A,
			0x0000022E,
			0x00000232,
			0x00000235,
			0x00000239,
			0x0000023D,
			0x00000241,
			0x00000245,
			0x00000248,
			0x0000024C,
			0x00000250,
			0x00000253,
			0x00000257,
			0x0000025B,
			0x0000025E,
			0x00000262,
			0x00000265,
			0x00000269,
			0x0000026C,
			0x0000026F,
			0x00000273,
			0x00000276,
			0x00000279,
			0x0000027D,
			0x00000280,
			0x00000283,
			0x00000286,
			0x00000289,
			0x0000028D,
			0x00000290,
			0x00000293,
			0x00000296,
			0x00000299,
			0x0000029C,
			0x0000029F,
			0x000002A2,
			0x000002A5,
			0x000002A8,
			0x000002AB,
			0x000002AE,
			0x000002B1,
			0x000002B4,
			0x000002B7,
			0x000002B9,
			0x000002BC,
			0x000002BF,
			0x000002C2,
			0x000002C5,
			0x000002C7,
			0x000002CA,
			0x000002CD,
			0x000002D0,
			0x000002D2,
			0x000002D5,
			0x000002D8,
			0x000002DA,
			0x000002DD,
			0x000002E0,
			0x000002E2,
			0x000002E5,
			0x000002E8,
			0x000002EA,
			0x000002ED,
			0x000002EF,
			0x000002F2,
			0x000002F4,
			0x000002F7,
			0x000002F9,
			0x000002FC,
			0x000002FE,
			0x00000301,
			0x00000303,
			0x00000306,
			0x00000308,
			0x0000030A,
			0x0000030D,
			0x0000030F,
			0x00000312,
			0x00000314,
			0x00000316,
			0x00000319,
			0x0000031B,
			0x0000031D,
			0x00000320,
			0x00000322,
			0x00000324,
			0x00000327,
			0x00000329,
			0x0000032B,
			0x0000032D,
			0x00000330,
			0x00000332,
			0x00000334,
			0x00000336,
			0x00000339,
			0x0000033B,
			0x0000033D,
			0x0000033F,
			0x00000341,
			0x00000344,
			0x00000346,
			0x00000348,
			0x0000034A,
			0x0000034C,
			0x0000034E,
			0x00000351,
			0x00000353,
			0x00000355,
			0x00000357,
			0x00000359,
			0x0000035B,
			0x0000035D,
			0x0000035F,
			0x00000361,
			0x00000363,
			0x00000365,
			0x00000367,
			0x00000369,
			0x0000036B,
			0x0000036D,
			0x0000036F,
			0x00000371,
			0x00000373,
			0x00000375,
			0x00000377,
			0x00000379,
			0x0000037B,
			0x0000037D,
			0x0000037F,
			0x00000381,
			0x00000383,
			0x00000385,
			0x00000387,
			0x00000389,
			0x0000038B,
			0x0000038D,
			0x0000038F,
			0x00000391,
			0x00000393,
			0x00000394,
			0x00000396,
			0x00000398,
			0x0000039A,
			0x0000039C,
			0x0000039E,
			0x000003A0,
			0x000003A2,
			0x000003A3,
			0x000003A5,
			0x000003A7,
			0x000003A9,
			0x000003AB,
			0x000003AD,
			0x000003AE,
			0x000003B0,
			0x000003B2,
			0x000003B4,
			0x000003B6,
			0x000003B7,
			0x000003B9,
			0x000003BB,
			0x000003BD,
			0x000003BE,
			0x000003C0,
			0x000003C2,
			0x000003C4,
			0x000003C6,
			0x000003C7,
			0x000003C9,
			0x000003CB,
			0x000003CC,
			0x000003CE,
			0x000003D0,
			0x000003D2,
			0x000003D3,
			0x000003D5,
			0x000003D7,
			0x000003D8,
			0x000003DA,
			0x000003DC,
			0x000003DE,
			0x000003DF,
			0x000003E1,
			0x000003E3,
			0x000003E4,
			0x000003E6,
			0x000003E8,
			0x000003E9,
			0x000003EB,
			0x000003ED,
			0x000003EE,
			0x000003F0,
			0x000003F1,
			0x000003F3,
			0x000003F5,
			0x000003F6,
			0x000003F8,
			0x000003FA,
			0x000003FB,
			0x000003FD,
			0x000003FE,
			0x000003FF,
		},
		// curve_ratio = 0.5
		{
			0x00000000,
			0x00000040,
			0x0000005B,
			0x0000006F,
			0x00000080,
			0x0000008F,
			0x0000009D,
			0x000000A9,
			0x000000B5,
			0x000000C0,
			0x000000CA,
			0x000000D4,
			0x000000DE,
			0x000000E7,
			0x000000EF,
			0x000000F8,
			0x00000100,
			0x00000108,
			0x00000110,
			0x00000117,
			0x0000011E,
			0x00000125,
			0x0000012C,
			0x00000133,
			0x0000013A,
			0x00000140,
			0x00000146,
			0x0000014D,
			0x00000153,
			0x00000159,
			0x0000015F,
			0x00000164,
			0x0000016A,
			0x00000170,
			0x00000175,
			0x0000017B,
			0x00000180,
			0x00000185,
			0x0000018B,
			0x00000190,
			0x00000195,
			0x0000019A,
			0x0000019F,
			0x000001A4,
			0x000001A9,
			0x000001AD,
			0x000001B2,
			0x000001B7,
			0x000001BB,
			0x000001C0,
			0x000001C5,
			0x000001C9,
			0x000001CE,
			0x000001D2,
			0x000001D6,
			0x000001DB,
			0x000001DF,
			0x000001E3,
			0x000001E7,
			0x000001EC,
			0x000001F0,
			0x000001F4,
			0x000001F8,
			0x000001FC,
			0x00000200,
			0x00000204,
			0x00000208,
			0x0000020C,
			0x00000210,
			0x00000214,
			0x00000217,
			0x0000021B,
			0x0000021F,
			0x00000223,
			0x00000227,
			0x0000022A,
			0x0000022E,
			0x00000232,
			0x00000235,
			0x00000239,
			0x0000023C,
			0x00000240,
			0x00000244,
			0x00000247,
			0x0000024B,
			0x0000024E,
			0x00000252,
			0x00000255,
			0x00000258,
			0x0000025C,
			0x0000025F,
			0x00000263,
			0x00000266,
			0x00000269,
			0x0000026D,
			0x00000270,
			0x00000273,
			0x00000276,
			0x0000027A,
			0x0000027D,
			0x00000280,
			0x00000283,
			0x00000286,
			0x0000028A,
			0x0000028D,
			0x00000290,
			0x00000293,
			0x00000296,
			0x00000299,
			0x0000029C,
			0x0000029F,
			0x000002A2,
			0x000002A5,
			0x000002A8,
			0x000002AB,
			0x000002AE,
			0x000002B1,
			0x000002B4,
			0x000002B7,
			0x000002BA,
			0x000002BD,
			0x000002C0,
			0x000002C3,
			0x000002C6,
			0x000002C9,
			0x000002CC,
			0x000002CE,
			0x000002D1,
			0x000002D4,
			0x000002D7,
			0x000002DA,
			0x000002DD,
			0x000002DF,
			0x000002E2,
			0x000002E5,
			0x000002E8,
			0x000002EA,
			0x000002ED,
			0x000002F0,
			0x000002F3,
			0x000002F5,
			0x000002F8,
			0x000002FB,
			0x000002FD,
			0x00000300,
			0x00000303,
			0x00000305,
			0x00000308,
			0x0000030B,
			0x0000030D,
			0x00000310,
			0x00000312,
			0x00000315,
			0x00000318,
			0x0000031A,
			0x0000031D,
			0x0000031F,
			0x00000322,
			0x00000324,
			0x00000327,
			0x0000032A,
			0x0000032C,
			0x0000032F,
			0x00000331,
			0x00000334,
			0x00000336,
			0x00000339,
			0x0000033B,
			0x0000033E,
			0x00000340,
			0x00000342,
			0x00000345,
			0x00000347,
			0x0000034A,
			0x0000034C,
			0x0000034F,
			0x00000351,
			0x00000353,
			0x00000356,
			0x00000358,
			0x0000035B,
			0x0000035D,
			0x0000035F,
			0x00000362,
			0x00000364,
			0x00000366,
			0x00000369,
			0x0000036B,
			0x0000036E,
			0x00000370,
			0x00000372,
			0x00000374,
			0x00000377,
			0x00000379,
			0x0000037B,
			0x0000037E,
			0x00000380,
			0x00000382,
			0x00000385,
			0x00000387,
			0x00000389,
			0x0000038B,
			0x0000038E,
			0x00000390,
			0x00000392,
			0x00000394,
			0x00000397,
			0x00000399,
			0x0000039B,
			0x0000039D,
			0x0000039F,
			0x000003A2,
			0x000003A4,
			0x000003A6,
			0x000003A8,
			0x000003AA,
			0x000003AD,
			0x000003AF,
			0x000003B1,
			0x000003B3,
			0x000003B5,
			0x000003B7,
			0x000003BA,
			0x000003BC,
			0x000003BE,
			0x000003C0,
			0x000003C2,
			0x000003C4,
			0x000003C6,
			0x000003C8,
			0x000003CB,
			0x000003CD,
			0x000003CF,
			0x000003D1,
			0x000003D3,
			0x000003D5,
			0x000003D7,
			0x000003D9,
			0x000003DB,
			0x000003DD,
			0x000003DF,
			0x000003E2,
			0x000003E4,
			0x000003E6,
			0x000003E8,
			0x000003EA,
			0x000003EC,
			0x000003EE,
			0x000003F0,
			0x000003F2,
			0x000003F4,
			0x000003F6,
			0x000003F8,
			0x000003FA,
			0x000003FC,
			0x000003FE,
			0x000003FF,
		},
		// curve_ratio = 0.6
		{
			0x00000000,
			0x00000025,
			0x00000038,
			0x00000047,
			0x00000054,
			0x00000061,
			0x0000006C,
			0x00000076,
			0x00000080,
			0x00000089,
			0x00000092,
			0x0000009B,
			0x000000A3,
			0x000000AB,
			0x000000B3,
			0x000000BB,
			0x000000C2,
			0x000000C9,
			0x000000D0,
			0x000000D7,
			0x000000DE,
			0x000000E4,
			0x000000EB,
			0x000000F1,
			0x000000F7,
			0x000000FE,
			0x00000104,
			0x0000010A,
			0x0000010F,
			0x00000115,
			0x0000011B,
			0x00000121,
			0x00000126,
			0x0000012C,
			0x00000131,
			0x00000136,
			0x0000013C,
			0x00000141,
			0x00000146,
			0x0000014B,
			0x00000150,
			0x00000155,
			0x0000015A,
			0x0000015F,
			0x00000164,
			0x00000169,
			0x0000016E,
			0x00000172,
			0x00000177,
			0x0000017C,
			0x00000180,
			0x00000185,
			0x0000018A,
			0x0000018E,
			0x00000193,
			0x00000197,
			0x0000019B,
			0x000001A0,
			0x000001A4,
			0x000001A8,
			0x000001AD,
			0x000001B1,
			0x000001B5,
			0x000001BA,
			0x000001BE,
			0x000001C2,
			0x000001C6,
			0x000001CA,
			0x000001CE,
			0x000001D2,
			0x000001D6,
			0x000001DA,
			0x000001DE,
			0x000001E2,
			0x000001E6,
			0x000001EA,
			0x000001EE,
			0x000001F2,
			0x000001F6,
			0x000001FA,
			0x000001FE,
			0x00000201,
			0x00000205,
			0x00000209,
			0x0000020D,
			0x00000210,
			0x00000214,
			0x00000218,
			0x0000021C,
			0x0000021F,
			0x00000223,
			0x00000227,
			0x0000022A,
			0x0000022E,
			0x00000231,
			0x00000235,
			0x00000238,
			0x0000023C,
			0x00000240,
			0x00000243,
			0x00000247,
			0x0000024A,
			0x0000024E,
			0x00000251,
			0x00000254,
			0x00000258,
			0x0000025B,
			0x0000025F,
			0x00000262,
			0x00000265,
			0x00000269,
			0x0000026C,
			0x00000270,
			0x00000273,
			0x00000276,
			0x0000027A,
			0x0000027D,
			0x00000280,
			0x00000283,
			0x00000287,
			0x0000028A,
			0x0000028D,
			0x00000290,
			0x00000294,
			0x00000297,
			0x0000029A,
			0x0000029D,
			0x000002A0,
			0x000002A4,
			0x000002A7,
			0x000002AA,
			0x000002AD,
			0x000002B0,
			0x000002B3,
			0x000002B6,
			0x000002BA,
			0x000002BD,
			0x000002C0,
			0x000002C3,
			0x000002C6,
			0x000002C9,
			0x000002CC,
			0x000002CF,
			0x000002D2,
			0x000002D5,
			0x000002D8,
			0x000002DB,
			0x000002DE,
			0x000002E1,
			0x000002E4,
			0x000002E7,
			0x000002EA,
			0x000002ED,
			0x000002F0,
			0x000002F3,
			0x000002F6,
			0x000002F9,
			0x000002FC,
			0x000002FF,
			0x00000301,
			0x00000304,
			0x00000307,
			0x0000030A,
			0x0000030D,
			0x00000310,
			0x00000313,
			0x00000316,
			0x00000318,
			0x0000031B,
			0x0000031E,
			0x00000321,
			0x00000324,
			0x00000327,
			0x00000329,
			0x0000032C,
			0x0000032F,
			0x00000332,
			0x00000335,
			0x00000337,
			0x0000033A,
			0x0000033D,
			0x00000340,
			0x00000342,
			0x00000345,
			0x00000348,
			0x0000034B,
			0x0000034D,
			0x00000350,
			0x00000353,
			0x00000356,
			0x00000358,
			0x0000035B,
			0x0000035E,
			0x00000360,
			0x00000363,
			0x00000366,
			0x00000368,
			0x0000036B,
			0x0000036E,
			0x00000370,
			0x00000373,
			0x00000376,
			0x00000378,
			0x0000037B,
			0x0000037E,
			0x00000380,
			0x00000383,
			0x00000385,
			0x00000388,
			0x0000038B,
			0x0000038D,
			0x00000390,
			0x00000392,
			0x00000395,
			0x00000398,
			0x0000039A,
			0x0000039D,
			0x0000039F,
			0x000003A2,
			0x000003A4,
			0x000003A7,
			0x000003AA,
			0x000003AC,
			0x000003AF,
			0x000003B1,
			0x000003B4,
			0x000003B6,
			0x000003B9,
			0x000003BB,
			0x000003BE,
			0x000003C0,
			0x000003C3,
			0x000003C5,
			0x000003C8,
			0x000003CA,
			0x000003CD,
			0x000003CF,
			0x000003D2,
			0x000003D4,
			0x000003D7,
			0x000003D9,
			0x000003DC,
			0x000003DE,
			0x000003E0,
			0x000003E3,
			0x000003E5,
			0x000003E8,
			0x000003EA,
			0x000003ED,
			0x000003EF,
			0x000003F2,
			0x000003F4,
			0x000003F6,
			0x000003F9,
			0x000003FB,
			0x000003FE,
			0x000003FF,
		},
		// curve_ratio = 0.7
		{
			0x00000000,
			0x00000015,
			0x00000022,
			0x0000002E,
			0x00000038,
			0x00000041,
			0x0000004A,
			0x00000052,
			0x0000005B,
			0x00000062,
			0x0000006A,
			0x00000071,
			0x00000078,
			0x0000007F,
			0x00000086,
			0x0000008D,
			0x00000093,
			0x00000099,
			0x000000A0,
			0x000000A6,
			0x000000AC,
			0x000000B2,
			0x000000B8,
			0x000000BE,
			0x000000C3,
			0x000000C9,
			0x000000CF,
			0x000000D4,
			0x000000DA,
			0x000000DF,
			0x000000E4,
			0x000000EA,
			0x000000EF,
			0x000000F4,
			0x000000F9,
			0x000000FE,
			0x00000103,
			0x00000108,
			0x0000010D,
			0x00000112,
			0x00000117,
			0x0000011C,
			0x00000121,
			0x00000126,
			0x0000012B,
			0x0000012F,
			0x00000134,
			0x00000139,
			0x0000013D,
			0x00000142,
			0x00000146,
			0x0000014B,
			0x00000150,
			0x00000154,
			0x00000159,
			0x0000015D,
			0x00000161,
			0x00000166,
			0x0000016A,
			0x0000016F,
			0x00000173,
			0x00000177,
			0x0000017B,
			0x00000180,
			0x00000184,
			0x00000188,
			0x0000018C,
			0x00000191,
			0x00000195,
			0x00000199,
			0x0000019D,
			0x000001A1,
			0x000001A5,
			0x000001A9,
			0x000001AE,
			0x000001B2,
			0x000001B6,
			0x000001BA,
			0x000001BE,
			0x000001C2,
			0x000001C6,
			0x000001CA,
			0x000001CE,
			0x000001D1,
			0x000001D5,
			0x000001D9,
			0x000001DD,
			0x000001E1,
			0x000001E5,
			0x000001E9,
			0x000001ED,
			0x000001F0,
			0x000001F4,
			0x000001F8,
			0x000001FC,
			0x00000200,
			0x00000203,
			0x00000207,
			0x0000020B,
			0x0000020F,
			0x00000212,
			0x00000216,
			0x0000021A,
			0x0000021D,
			0x00000221,
			0x00000225,
			0x00000228,
			0x0000022C,
			0x00000230,
			0x00000233,
			0x00000237,
			0x0000023B,
			0x0000023E,
			0x00000242,
			0x00000245,
			0x00000249,
			0x0000024C,
			0x00000250,
			0x00000253,
			0x00000257,
			0x0000025B,
			0x0000025E,
			0x00000262,
			0x00000265,
			0x00000268,
			0x0000026C,
			0x0000026F,
			0x00000273,
			0x00000276,
			0x0000027A,
			0x0000027D,
			0x00000281,
			0x00000284,
			0x00000287,
			0x0000028B,
			0x0000028E,
			0x00000292,
			0x00000295,
			0x00000298,
			0x0000029C,
			0x0000029F,
			0x000002A3,
			0x000002A6,
			0x000002A9,
			0x000002AD,
			0x000002B0,
			0x000002B3,
			0x000002B6,
			0x000002BA,
			0x000002BD,
			0x000002C0,
			0x000002C4,
			0x000002C7,
			0x000002CA,
			0x000002CD,
			0x000002D1,
			0x000002D4,
			0x000002D7,
			0x000002DA,
			0x000002DE,
			0x000002E1,
			0x000002E4,
			0x000002E7,
			0x000002EB,
			0x000002EE,
			0x000002F1,
			0x000002F4,
			0x000002F7,
			0x000002FB,
			0x000002FE,
			0x00000301,
			0x00000304,
			0x00000307,
			0x0000030A,
			0x0000030D,
			0x00000311,
			0x00000314,
			0x00000317,
			0x0000031A,
			0x0000031D,
			0x00000320,
			0x00000323,
			0x00000326,
			0x0000032A,
			0x0000032D,
			0x00000330,
			0x00000333,
			0x00000336,
			0x00000339,
			0x0000033C,
			0x0000033F,
			0x00000342,
			0x00000345,
			0x00000348,
			0x0000034B,
			0x0000034E,
			0x00000351,
			0x00000354,
			0x00000357,
			0x0000035A,
			0x0000035D,
			0x00000361,
			0x00000364,
			0x00000367,
			0x0000036A,
			0x0000036D,
			0x00000370,
			0x00000372,
			0x00000375,
			0x00000378,
			0x0000037B,
			0x0000037E,
			0x00000381,
			0x00000384,
			0x00000387,
			0x0000038A,
			0x0000038D,
			0x00000390,
			0x00000393,
			0x00000396,
			0x00000399,
			0x0000039C,
			0x0000039F,
			0x000003A2,
			0x000003A5,
			0x000003A8,
			0x000003AA,
			0x000003AD,
			0x000003B0,
			0x000003B3,
			0x000003B6,
			0x000003B9,
			0x000003BC,
			0x000003BF,
			0x000003C2,
			0x000003C4,
			0x000003C7,
			0x000003CA,
			0x000003CD,
			0x000003D0,
			0x000003D3,
			0x000003D6,
			0x000003D8,
			0x000003DB,
			0x000003DE,
			0x000003E1,
			0x000003E4,
			0x000003E7,
			0x000003E9,
			0x000003EC,
			0x000003EF,
			0x000003F2,
			0x000003F5,
			0x000003F8,
			0x000003FA,
			0x000003FD,
			0x000003FF,
		},
		// curve_ratio = 0.8
		{
			0x00000000,
			0x0000000C,
			0x00000015,
			0x0000001D,
			0x00000025,
			0x0000002C,
			0x00000033,
			0x0000003A,
			0x00000040,
			0x00000046,
			0x0000004D,
			0x00000053,
			0x00000059,
			0x0000005E,
			0x00000064,
			0x0000006A,
			0x0000006F,
			0x00000075,
			0x0000007A,
			0x00000080,
			0x00000085,
			0x0000008B,
			0x00000090,
			0x00000095,
			0x0000009A,
			0x0000009F,
			0x000000A4,
			0x000000A9,
			0x000000AE,
			0x000000B3,
			0x000000B8,
			0x000000BD,
			0x000000C2,
			0x000000C7,
			0x000000CC,
			0x000000D0,
			0x000000D5,
			0x000000DA,
			0x000000DF,
			0x000000E3,
			0x000000E8,
			0x000000ED,
			0x000000F1,
			0x000000F6,
			0x000000FA,
			0x000000FF,
			0x00000103,
			0x00000108,
			0x0000010C,
			0x00000111,
			0x00000115,
			0x0000011A,
			0x0000011E,
			0x00000122,
			0x00000127,
			0x0000012B,
			0x00000130,
			0x00000134,
			0x00000138,
			0x0000013D,
			0x00000141,
			0x00000145,
			0x00000149,
			0x0000014E,
			0x00000152,
			0x00000156,
			0x0000015A,
			0x0000015E,
			0x00000163,
			0x00000167,
			0x0000016B,
			0x0000016F,
			0x00000173,
			0x00000177,
			0x0000017B,
			0x0000017F,
			0x00000184,
			0x00000188,
			0x0000018C,
			0x00000190,
			0x00000194,
			0x00000198,
			0x0000019C,
			0x000001A0,
			0x000001A4,
			0x000001A8,
			0x000001AC,
			0x000001B0,
			0x000001B4,
			0x000001B8,
			0x000001BC,
			0x000001C0,
			0x000001C4,
			0x000001C8,
			0x000001CB,
			0x000001CF,
			0x000001D3,
			0x000001D7,
			0x000001DB,
			0x000001DF,
			0x000001E3,
			0x000001E7,
			0x000001EA,
			0x000001EE,
			0x000001F2,
			0x000001F6,
			0x000001FA,
			0x000001FE,
			0x00000201,
			0x00000205,
			0x00000209,
			0x0000020D,
			0x00000211,
			0x00000214,
			0x00000218,
			0x0000021C,
			0x00000220,
			0x00000223,
			0x00000227,
			0x0000022B,
			0x0000022F,
			0x00000232,
			0x00000236,
			0x0000023A,
			0x0000023D,
			0x00000241,
			0x00000245,
			0x00000248,
			0x0000024C,
			0x00000250,
			0x00000253,
			0x00000257,
			0x0000025B,
			0x0000025E,
			0x00000262,
			0x00000266,
			0x00000269,
			0x0000026D,
			0x00000271,
			0x00000274,
			0x00000278,
			0x0000027B,
			0x0000027F,
			0x00000283,
			0x00000286,
			0x0000028A,
			0x0000028D,
			0x00000291,
			0x00000295,
			0x00000298,
			0x0000029C,
			0x0000029F,
			0x000002A3,
			0x000002A6,
			0x000002AA,
			0x000002AD,
			0x000002B1,
			0x000002B5,
			0x000002B8,
			0x000002BC,
			0x000002BF,
			0x000002C3,
			0x000002C6,
			0x000002CA,
			0x000002CD,
			0x000002D1,
			0x000002D4,
			0x000002D8,
			0x000002DB,
			0x000002DF,
			0x000002E2,
			0x000002E5,
			0x000002E9,
			0x000002EC,
			0x000002F0,
			0x000002F3,
			0x000002F7,
			0x000002FA,
			0x000002FE,
			0x00000301,
			0x00000305,
			0x00000308,
			0x0000030B,
			0x0000030F,
			0x00000312,
			0x00000316,
			0x00000319,
			0x0000031C,
			0x00000320,
			0x00000323,
			0x00000327,
			0x0000032A,
			0x0000032D,
			0x00000331,
			0x00000334,
			0x00000338,
			0x0000033B,
			0x0000033E,
			0x00000342,
			0x00000345,
			0x00000348,
			0x0000034C,
			0x0000034F,
			0x00000353,
			0x00000356,
			0x00000359,
			0x0000035D,
			0x00000360,
			0x00000363,
			0x00000367,
			0x0000036A,
			0x0000036D,
			0x00000371,
			0x00000374,
			0x00000377,
			0x0000037B,
			0x0000037E,
			0x00000381,
			0x00000384,
			0x00000388,
			0x0000038B,
			0x0000038E,
			0x00000392,
			0x00000395,
			0x00000398,
			0x0000039C,
			0x0000039F,
			0x000003A2,
			0x000003A5,
			0x000003A9,
			0x000003AC,
			0x000003AF,
			0x000003B2,
			0x000003B6,
			0x000003B9,
			0x000003BC,
			0x000003BF,
			0x000003C3,
			0x000003C6,
			0x000003C9,
			0x000003CC,
			0x000003D0,
			0x000003D3,
			0x000003D6,
			0x000003D9,
			0x000003DD,
			0x000003E0,
			0x000003E3,
			0x000003E6,
			0x000003EA,
			0x000003ED,
			0x000003F0,
			0x000003F3,
			0x000003F6,
			0x000003FA,
			0x000003FD,
			0x000003FF,
		},
		// curve_ratio = 0.9
		{
			0x00000000,
			0x00000007,
			0x0000000D,
			0x00000013,
			0x00000018,
			0x0000001E,
			0x00000023,
			0x00000028,
			0x0000002D,
			0x00000032,
			0x00000037,
			0x0000003C,
			0x00000041,
			0x00000046,
			0x0000004B,
			0x00000050,
			0x00000054,
			0x00000059,
			0x0000005E,
			0x00000063,
			0x00000067,
			0x0000006C,
			0x00000070,
			0x00000075,
			0x0000007A,
			0x0000007E,
			0x00000083,
			0x00000087,
			0x0000008C,
			0x00000090,
			0x00000095,
			0x00000099,
			0x0000009E,
			0x000000A2,
			0x000000A6,
			0x000000AB,
			0x000000AF,
			0x000000B4,
			0x000000B8,
			0x000000BC,
			0x000000C1,
			0x000000C5,
			0x000000C9,
			0x000000CE,
			0x000000D2,
			0x000000D6,
			0x000000DA,
			0x000000DF,
			0x000000E3,
			0x000000E7,
			0x000000EB,
			0x000000F0,
			0x000000F4,
			0x000000F8,
			0x000000FC,
			0x00000101,
			0x00000105,
			0x00000109,
			0x0000010D,
			0x00000111,
			0x00000115,
			0x0000011A,
			0x0000011E,
			0x00000122,
			0x00000126,
			0x0000012A,
			0x0000012E,
			0x00000132,
			0x00000137,
			0x0000013B,
			0x0000013F,
			0x00000143,
			0x00000147,
			0x0000014B,
			0x0000014F,
			0x00000153,
			0x00000157,
			0x0000015B,
			0x0000015F,
			0x00000163,
			0x00000167,
			0x0000016C,
			0x00000170,
			0x00000174,
			0x00000178,
			0x0000017C,
			0x00000180,
			0x00000184,
			0x00000188,
			0x0000018C,
			0x00000190,
			0x00000194,
			0x00000198,
			0x0000019C,
			0x000001A0,
			0x000001A4,
			0x000001A8,
			0x000001AC,
			0x000001B0,
			0x000001B3,
			0x000001B7,
			0x000001BB,
			0x000001BF,
			0x000001C3,
			0x000001C7,
			0x000001CB,
			0x000001CF,
			0x000001D3,
			0x000001D7,
			0x000001DB,
			0x000001DF,
			0x000001E3,
			0x000001E7,
			0x000001EB,
			0x000001EE,
			0x000001F2,
			0x000001F6,
			0x000001FA,
			0x000001FE,
			0x00000202,
			0x00000206,
			0x0000020A,
			0x0000020E,
			0x00000211,
			0x00000215,
			0x00000219,
			0x0000021D,
			0x00000221,
			0x00000225,
			0x00000229,
			0x0000022C,
			0x00000230,
			0x00000234,
			0x00000238,
			0x0000023C,
			0x00000240,
			0x00000244,
			0x00000247,
			0x0000024B,
			0x0000024F,
			0x00000253,
			0x00000257,
			0x0000025A,
			0x0000025E,
			0x00000262,
			0x00000266,
			0x0000026A,
			0x0000026E,
			0x00000271,
			0x00000275,
			0x00000279,
			0x0000027D,
			0x00000281,
			0x00000284,
			0x00000288,
			0x0000028C,
			0x00000290,
			0x00000293,
			0x00000297,
			0x0000029B,
			0x0000029F,
			0x000002A3,
			0x000002A6,
			0x000002AA,
			0x000002AE,
			0x000002B2,
			0x000002B5,
			0x000002B9,
			0x000002BD,
			0x000002C1,
			0x000002C4,
			0x000002C8,
			0x000002CC,
			0x000002D0,
			0x000002D3,
			0x000002D7,
			0x000002DB,
			0x000002DF,
			0x000002E2,
			0x000002E6,
			0x000002EA,
			0x000002EE,
			0x000002F1,
			0x000002F5,
			0x000002F9,
			0x000002FC,
			0x00000300,
			0x00000304,
			0x00000308,
			0x0000030B,
			0x0000030F,
			0x00000313,
			0x00000316,
			0x0000031A,
			0x0000031E,
			0x00000322,
			0x00000325,
			0x00000329,
			0x0000032D,
			0x00000330,
			0x00000334,
			0x00000338,
			0x0000033B,
			0x0000033F,
			0x00000343,
			0x00000346,
			0x0000034A,
			0x0000034E,
			0x00000351,
			0x00000355,
			0x00000359,
			0x0000035C,
			0x00000360,
			0x00000364,
			0x00000367,
			0x0000036B,
			0x0000036F,
			0x00000372,
			0x00000376,
			0x0000037A,
			0x0000037D,
			0x00000381,
			0x00000385,
			0x00000388,
			0x0000038C,
			0x00000390,
			0x00000393,
			0x00000397,
			0x0000039B,
			0x0000039E,
			0x000003A2,
			0x000003A6,
			0x000003A9,
			0x000003AD,
			0x000003B0,
			0x000003B4,
			0x000003B8,
			0x000003BB,
			0x000003BF,
			0x000003C3,
			0x000003C6,
			0x000003CA,
			0x000003CD,
			0x000003D1,
			0x000003D5,
			0x000003D8,
			0x000003DC,
			0x000003E0,
			0x000003E3,
			0x000003E7,
			0x000003EA,
			0x000003EE,
			0x000003F2,
			0x000003F5,
			0x000003F9,
			0x000003FC,
			0x000003FF,
		},
	};

// echo 0x00000rgb > .../gamma_table_curve_ratio
// curve_ratio_r/g/b: '0' for straight, '1' for ^0.1, ..., '9' for ^0.9

bool set_gamma_table_with_curve_ratio(unsigned int r,unsigned int g,unsigned int b)
{


    bool flag = false;
	unsigned int i = 0;

    // parameter validation check
    if ((r>9) || (g>9) || (b>9))
    {
        return false;
    }

    // store gamma table enable/disable status
    flag = (RD(0x2d30)&0x10000000)?true:false;

    // gamma table disable, gamma table vbus mode
    WR(0x2d30, 0x00000001);

    // point to gamma_r
    WR(0x2dc2, 0x00000000);
    // write gamma_r
    for (i = 0; i < 257; i++)
    {
        WR(0x2dc3, gamma[r][i]);
    }

    // point to gamma_r
    WR(0x2dc2, 0x00000000);
    // varify gamma_r
    for (i = 0; i < 257; i++)
    {
        if (gamma[r][i] != RD(0x2dc3))
        {
            // retrieve gamma table enable/disable status, gamma table hardware mode
            WR(0x2d30, (flag?0x10000000:0x00000000));
            // return with failure
            return false;
        }
    }

    // point to gamma_g
    WR(0x2dc2, 0x00000200);
    // write gamma_g
    for (i = 0; i < 257; i++)
    {
        WR(0x2dc3, gamma[g][i]);
    }

    // point to gamma_g
    WR(0x2dc2, 0x00000200);
    // varify gamma_g
    for (i = 0; i < 257; i++)
    {
        if (gamma[g][i] != RD(0x2dc3))
        {
            // retrieve gamma table enable/disable status, gamma table hardware mode
            WR(0x2d30, (flag?0x10000000:0x00000000));
            // return with failure
            return false;
        }
    }

    // point to gamma_b
    WR(0x2dc2, 0x00000400);
    // write gamma_b
    for (i = 0; i < 257; i++)
    {
        WR(0x2dc3, gamma[b][i]);
    }

    // point to gamma_b
    WR(0x2dc2, 0x00000400);
    // varify gamma_b
    for (i = 0; i < 257; i++)
    {
        if (gamma[b][i] != RD(0x2dc3))
        {
            // retrieve gamma table enable/disable status, gamma table hardware mode
            WR(0x2d30, (flag?0x10000000:0x00000000));
            // return with failure
            return false;
        }
    }

    // retrieve gamma table enable/disable status, gamma table hardware mode
    WR(0x2d30, (flag?0x10000000:0x00000000));

    // return with success
    return true;
}


